use strict;
use Test::Simple tests => 20;

use DE_EPAGES::Core::API::Error  qw ( GetError ExistsError );
use DE_EPAGES::Core::API::File   qw ( CopyFile );
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Object::API::XMLCompare qw ( CompareXML );
use DE_EPAGES::Installer::API::CartridgesXMLFilePatcher
    qw( ReadCartridgesXmlFile GetCartridgeDependencies CartridgeDependencyError );
use DE_EPAGES::Object::API::XMLStructureImportHandler;

sub Main {

    local $| = 1;

    my $time1 = time;

    my $hOptions = {'checkdelete' => '1', 'checkreference' => '1', 'printmessage' => '1'};
    my ($CartridgesXml, $FilePatcher);

    # test cartridges

    ($CartridgesXml, $FilePatcher) = _getCartridgesXml(1, 'imp');
    $FilePatcher->importElements($CartridgesXml->{changes});
    ok (CompareXML($CartridgesXml->{result}, $CartridgesXml->{expected}, $hOptions), 'import (1.1)');
    ok (CompareXML($CartridgesXml->{expected}, $CartridgesXml->{result}, $hOptions), 'import (1.2)');

    ($CartridgesXml, $FilePatcher) = _getCartridgesXml(2, 'del');
    $FilePatcher->deleteElements($CartridgesXml->{changes});
    ok (CompareXML($CartridgesXml->{result}, $CartridgesXml->{expected}, $hOptions), 'delete (2.1)');
    ok (CompareXML($CartridgesXml->{expected}, $CartridgesXml->{result}, $hOptions), 'delete (2.2)');

    # test cartridge positions

    ($CartridgesXml, $FilePatcher) = _getCartridgesXml(1);
    my $XMLStructImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
    my $XMLStruct = $XMLStructImportHandler->getXMLStructure(
        $CartridgesXml->{expected}, {'XML-Type' => 'epages_cartridges_xml'});

    my $ExternalShipping = 'Cartridge{Package->DE_EPAGES::ExternalShipping}';
    my $MRW = 'Cartridge{Package->DE_EPAGES::MRW}';

    my $Position1 = $XMLStruct->getXMLStructAttribute("/DatabaseType{Name->Store}/$ExternalShipping",'Position');
    my $Position2 = $XMLStruct->getXMLStructAttribute("/DatabaseType{Name->Store}/$MRW",'Position');
    ok(($XMLStruct->existsElement("/DatabaseType{Name->Store}/$MRW") and $Position1 < $Position2),
       'DE_EPAGES::ExternalShipping is before DE_EPAGES::MRW in Store'
    );

    $Position1 = $XMLStruct->getXMLStructAttribute("/DatabaseType{Name->Fantasy}/$ExternalShipping",'Position');
    $Position2 = $XMLStruct->getXMLStructAttribute("/DatabaseType{Name->Fantasy}/$MRW",'Position');
    ok(($XMLStruct->existsElement("/DatabaseType{Name->Fantasy}/$MRW") and $Position1 < $Position2),
       'DE_EPAGES::ExternalShipping is before DE_EPAGES::MRW in Fantasy'
    );

    $Position1 = $XMLStruct->getXMLStructAttribute("/DatabaseType{Name->Fantasy3}/$ExternalShipping",'Position');
    $Position2 = $XMLStruct->getXMLStructAttribute("/DatabaseType{Name->Fantasy3}/$MRW",'Position');
    ok((not $XMLStruct->existsElement("/DatabaseType{Name->Fantasy3}/$MRW")),
       'DE_EPAGES::MRW is not in Fantasy3' );

    # test custom storetypes

    ($CartridgesXml, $FilePatcher) = _getCartridgesXml(3, 'del');
    $FilePatcher->deleteElements($CartridgesXml->{changes});
    ok (CompareXML($CartridgesXml->{result}, $CartridgesXml->{expected}, $hOptions), 'delete custom (3.1)');
    ok (CompareXML($CartridgesXml->{expected}, $CartridgesXml->{result}, $hOptions), 'delete custom (3.2)');

    ($CartridgesXml, $FilePatcher) = _getCartridgesXml(4, 'del');
    $FilePatcher->deleteElements($CartridgesXml->{changes});
    ok (CompareXML($CartridgesXml->{result}, $CartridgesXml->{expected}, $hOptions), 'delete custom (4.1)');
    ok (CompareXML($CartridgesXml->{expected}, $CartridgesXml->{result}, $hOptions), 'delete custom (4.2)');

    # test missing independent cartridge

    ($CartridgesXml, $FilePatcher) = _getCartridgesXml(1);
    my $hCartridgesXml = ReadCartridgesXmlFile($CartridgesXml->{expected});
    my ($DependentCartridges) = GetCartridgeDependencies($hCartridgesXml);

    eval { CartridgeDependencyError($DependentCartridges); };
    my $Cartridge = ExistsError() ? GetError()->{Vars}{Cartridge} : undef;
    my $Requires = ExistsError() ? GetError()->{Vars}{Requires} : undef;
    ok((ExistsError() and GetError()->code eq 'DEPENDENCY_MISSING' and
        $Cartridge eq 'DE_EPAGES::JQuery' and $Requires eq 'DE_EPAGES::Core'),
       'DE_EPAGES::JQuery requires DE_EPAGES::Core (missing)'
    );

    # test default error (no other error found)

    $DependentCartridges->{'DE_EPAGES::Core'} = {};
    eval { CartridgeDependencyError($DependentCartridges); };
    ok((ExistsError() and GetError()->code eq 'DEPENDENCIES_NOT_RESOLVABLE'),
       'DE_EPAGES::JQuery requires DE_EPAGES::Core (exists)'
    );

    # test cartridge depending on itself

    $DependentCartridges->{'DE_EPAGES::Order'}{'DE_EPAGES::Order'} = 1;

    eval { CartridgeDependencyError($DependentCartridges); };
    my $CircularDependency = ExistsError() ? GetError()->{Vars}{CircularDependency} : undef;
    my $ExpectedDependency = 'DE_EPAGES::Order->DE_EPAGES::Order';
    ok((ExistsError() and GetError()->code eq 'CIRCULAR_DEPENDENCY' and
        $CircularDependency eq $ExpectedDependency),
       "circular dependency: $ExpectedDependency"
    );

    # test direct circular dependency

    delete $DependentCartridges->{'DE_EPAGES::Order'}{'DE_EPAGES::Order'};
    $DependentCartridges->{'DE_EPAGES::Dictionary'}{'DE_EPAGES::Object'} = 1;

    eval { CartridgeDependencyError($DependentCartridges); };
    $CircularDependency = ExistsError() ? GetError()->{Vars}{CircularDependency} : undef;
    $ExpectedDependency = 'DE_EPAGES::Dictionary->DE_EPAGES::Object->DE_EPAGES::Dictionary';
    ok((ExistsError() and GetError()->code eq 'CIRCULAR_DEPENDENCY' and
        $CircularDependency eq $ExpectedDependency),
       "circular dependency: $ExpectedDependency"
    );

    # test indirect circular dependency

    delete $DependentCartridges->{'DE_EPAGES::Dictionary'}{'DE_EPAGES::Object'};
    $DependentCartridges->{'DE_EPAGES::Order'}{'DE_EPAGES::PayPalPro'} = 1;

    eval { CartridgeDependencyError($DependentCartridges); };
    $CircularDependency = ExistsError() ? GetError()->{Vars}{CircularDependency} : undef;
    $ExpectedDependency = 'DE_EPAGES::ExternalPayment->DE_EPAGES::Order->...->DE_EPAGES::ExternalPayment';
    ok((ExistsError() and GetError()->code eq 'CIRCULAR_DEPENDENCY' and
        $CircularDependency eq $ExpectedDependency),
       "circular dependency: $ExpectedDependency"
    );

    # test corrupt Cartridges.xml

    ($CartridgesXml, $FilePatcher) = _getCartridgesXml(5);
    eval { $FilePatcher->importElements(undef, {check_dependencies => 1}); };
    $Cartridge = ExistsError() ? GetError()->{Vars}{Cartridge} : undef;
    $Requires = ExistsError() ? GetError()->{Vars}{Requires} : undef;
    ok((ExistsError() and GetError()->code eq 'DEPENDENCY_MISSING' and
        $Cartridge eq 'DE_EPAGES::MultiStore' and $Requires eq 'DE_EPAGES::MultiStoreCommon'),
       'DE_EPAGES::MultiStore requires DE_EPAGES::MultiStoreCommon (missing)'
    );

    eval { $FilePatcher->importElements(); };
    ok((not ExistsError()),
       'DE_EPAGES::MultiStore requires DE_EPAGES::MultiStoreCommon (ignored)'
    );
    ok (CompareXML($CartridgesXml->{result}, $CartridgesXml->{expected}, $hOptions), 'sort (5.1)');
    ok (CompareXML($CartridgesXml->{expected}, $CartridgesXml->{result}, $hOptions), 'sort (5.2)');

    my $time2 = time;
    printf "need %.3f seconds\n", $time2-$time1;
}

sub _getCartridgesXml {
    my ($Number, $Changes) = @_;

    my %CartridgesXml = (
        file => "Cartridges$Number.xml",
        changes => "${Changes}_Cartridges$Number.xml",
        result => "result_Cartridges$Number.xml",
        expected => "expected_Cartridges$Number.xml",
        );

    CopyFile($CartridgesXml{file}, $CartridgesXml{result});
    my $FilePatcher = DE_EPAGES::Installer::API::CartridgesXMLFilePatcher->new($CartridgesXml{result});

    return (\%CartridgesXml, $FilePatcher);
}

RunScript( Message => 'long', Sub => \&Main);

1;
